# pylint: disable=useless-import-alias
from ._impl import (
    CloseReason as CloseReason,
    ConnectionClosed as ConnectionClosed,
    ConnectionRejected as ConnectionRejected,
    ConnectionTimeout as ConnectionTimeout,
    connect_websocket as connect_websocket,
    connect_websocket_url as connect_websocket_url,
    DisconnectionTimeout as DisconnectionTimeout,
    Endpoint as Endpoint,
    HandshakeError as HandshakeError,
    open_websocket as open_websocket,
    open_websocket_url as open_websocket_url,
    WebSocketConnection as WebSocketConnection,
    WebSocketRequest as WebSocketRequest,
    WebSocketServer as WebSocketServer,
    wrap_client_stream as wrap_client_stream,
    wrap_server_stream as wrap_server_stream,
    serve_websocket as serve_websocket,
)
from ._version import __version__ as __version__



'''
This simple WebSocket server responds to text messages by reversing each
message string and sending it back.

It also handles ping/pong automatically and will correctly close down a
connection when the client requests it.

To use SSL/TLS: install the `trustme` package from PyPI and run the
`generate-cert.py` script in this directory.
'''
import argparse
import logging
import pathlib
import ssl

import trio
from trio_websocket import serve_websocket, ConnectionClosed, WebSocketRequest


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()
here = pathlib.Path(__file__).parent


def parse_args() -> argparse.Namespace:
    ''' Parse command line arguments. '''
    parser = argparse.ArgumentParser(description='Example trio-websocket client')
    parser.add_argument('--ssl', action='store_true', help='Use SSL')
    parser.add_argument('host', help='Host interface to bind. If omitted, '
        'then bind all interfaces.', nargs='?')
    parser.add_argument('port', type=int, help='Port to bind.')
    return parser.parse_args()


async def main(args: argparse.Namespace) -> None:
    ''' Main entry point. '''
    logging.info('Starting websocket server…')
    if args.ssl:
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        try:
            ssl_context.load_cert_chain(here / 'fake.server.pem')
        except FileNotFoundError:
            logging.error('Did not find file "fake.server.pem". You need to run'
                ' generate-cert.py')
    else:
        ssl_context = None
    host = None if args.host == '*' else args.host
    await serve_websocket(handler, host, args.port, ssl_context)


async def handler(request: WebSocketRequest) -> None:
    ''' Reverse incoming websocket messages and send them back. '''
    logging.info('Handler starting on path "%s"', request.path)
    ws = await request.accept()
    while True:
        try:
            message = await ws.get_message()
            await ws.send_message(message[::-1])
        except ConnectionClosed:
            logging.info('Connection closed')
            break
    logging.info('Handler exiting')


if __name__ == '__main__':
    try:
        trio.run(main, parse_args())
    except KeyboardInterrupt:
        print()



'''
This interactive WebSocket client allows the user to send frames to a WebSocket
server, including text message, ping, and close frames.

To use SSL/TLS: install the `trustme` package from PyPI and run the
`generate-cert.py` script in this directory.
'''
import argparse
import logging
import pathlib
import ssl
import sys
import urllib.parse
from typing import NoReturn

import trio
from trio_websocket import (
    open_websocket_url,
    ConnectionClosed,
    HandshakeError,
    WebSocketConnection,
    CloseReason,
)


logging.basicConfig(level=logging.DEBUG)
here = pathlib.Path(__file__).parent


def commands() -> None:
    ''' Print the supported commands. '''
    print('Commands: ')
    print('send <MESSAGE>   -> send message')
    print('ping <PAYLOAD>   -> send ping with payload')
    print('close [<REASON>] -> politely close connection with optional reason')
    print()


def parse_args() -> argparse.Namespace:
    ''' Parse command line arguments. '''
    parser = argparse.ArgumentParser(description='Example trio-websocket client')
    parser.add_argument('--heartbeat', action='store_true',
        help='Create a heartbeat task')
    parser.add_argument('url', help='WebSocket URL to connect to')
    return parser.parse_args()


async def main(args: argparse.Namespace) -> bool:
    ''' Main entry point, returning False in the case of logged error. '''
    if urllib.parse.urlsplit(args.url).scheme == 'wss':
        # Configure SSL context to handle our self-signed certificate. Most
        # clients won't need to do this.
        try:
            ssl_context = ssl.create_default_context()
            ssl_context.load_verify_locations(here / 'fake.ca.pem')
        except FileNotFoundError:
            logging.error('Did not find file "fake.ca.pem". You need to run'
                ' generate-cert.py')
            return False
    else:
        ssl_context = None
    try:
        logging.debug('Connecting to WebSocket…')
        async with open_websocket_url(args.url, ssl_context) as conn:
            await handle_connection(conn, args.heartbeat)
    except HandshakeError as e:
        logging.error('Connection attempt failed: %s', e)
        return False
    return True


async def handle_connection(ws: WebSocketConnection, use_heartbeat: bool) -> None:
    ''' Handle the connection. '''
    logging.debug('Connected!')
    try:
        async with trio.open_nursery() as nursery:
            if use_heartbeat:
                nursery.start_soon(heartbeat, ws, 1, 15)
            nursery.start_soon(get_commands, ws)
            nursery.start_soon(get_messages, ws)
    except ConnectionClosed as cc:
        assert isinstance(cc.reason, CloseReason)
        reason = '<no reason>' if cc.reason.reason is None else f'"{cc.reason.reason}"'
        print(f'Closed: {cc.reason.code}/{cc.reason.name} {reason}')


async def heartbeat(ws: WebSocketConnection, timeout: float, interval: float) -> NoReturn:
    '''
    Send periodic pings on WebSocket ``ws``.

    Wait up to ``timeout`` seconds to send a ping and receive a pong. Raises
    ``TooSlowError`` if the timeout is exceeded. If a pong is received, then
    wait ``interval`` seconds before sending the next ping.

    This function runs until cancelled.

    :param ws: A WebSocket to send heartbeat pings on.
    :param float timeout: Timeout in seconds.
    :param float interval: Interval between receiving pong and sending next
        ping, in seconds.
    :raises: ``ConnectionClosed`` if ``ws`` is closed.
    :raises: ``TooSlowError`` if the timeout expires.
    :returns: This function runs until cancelled.
    '''
    while True:
        with trio.fail_after(timeout):
            await ws.ping()
        await trio.sleep(interval)


async def get_commands(ws: WebSocketConnection) -> None:
    ''' In a loop: get a command from the user and execute it. '''
    while True:
        cmd = await trio.to_thread.run_sync(input, 'cmd> ')
        if cmd.startswith('ping'):
            payload = cmd[5:].encode('utf8') or None
            await ws.ping(payload)
        elif cmd.startswith('send'):
            message = cmd[5:] or None
            if message is None:
                logging.error('The "send" command requires a message.')
            else:
                await ws.send_message(message)
        elif cmd.startswith('close'):
            reason = cmd[6:] or None
            await ws.aclose(code=1000, reason=reason)
            break
        else:
            commands()
        # Allow time to receive response and log print logs:
        await trio.sleep(0.25)


async def get_messages(ws: WebSocketConnection) -> None:
    ''' In a loop: get a WebSocket message and print it out. '''
    while True:
        message = await ws.get_message()
        print(f'message: {message!r}')


if __name__ == '__main__':
    try:
        if not trio.run(main, parse_args()):
            sys.exit(1)
    except (KeyboardInterrupt, EOFError):
        print()



'''
This test client runs against the Autobahn test server. It is based on the
test_client.py in wsproto.
'''
import argparse
import json
import logging
import sys

import trio
from trio_websocket import open_websocket_url, ConnectionClosed


AGENT = 'trio-websocket'
MAX_MESSAGE_SIZE = 16 * 1024 * 1024
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('client')


async def get_case_count(url: str) -> int:
    url = url + '/getCaseCount'
    async with open_websocket_url(url) as conn:
        case_count = await conn.get_message()
        logger.info('Case count=%s', case_count)
    return int(case_count)


async def get_case_info(url: str, case: str) -> object:
    url = f'{url}/getCaseInfo?case={case}'
    async with open_websocket_url(url) as conn:
        return json.loads(await conn.get_message())


async def run_case(url: str, case: str) -> None:
    url = f'{url}/runCase?case={case}&agent={AGENT}'
    try:
        async with open_websocket_url(url, max_message_size=MAX_MESSAGE_SIZE) as conn:
            while True:
                data = await conn.get_message()
                await conn.send_message(data)
    except ConnectionClosed:
        pass


async def update_reports(url: str) -> None:
    url = f'{url}/updateReports?agent={AGENT}'
    async with open_websocket_url(url) as conn:
        # This command runs as soon as we connect to it, so we don't need to
        # send any messages.
        pass


async def run_tests(args: argparse.Namespace) -> None:
    logger = logging.getLogger('trio-websocket')
    if args.debug_cases:
        # Don't fetch case count when debugging a subset of test cases. It adds
        # noise to the debug logging.
        case_count = None
        test_cases = args.debug_cases
    else:
        case_count = await get_case_count(args.url)
        test_cases = list(range(1, case_count + 1))
    exception_cases = []
    for case in test_cases:
        result = await get_case_info(args.url, case)
        assert isinstance(result, dict)
        case_id = result['id']
        assert isinstance(case_id, int)
        if case_count:
            logger.info("Running test case %s (%d of %d)", case_id, case, case_count)
        else:
            logger.info("Debugging test case %s (%d)", case_id, case)
            logger.setLevel(logging.DEBUG)
        try:
            await run_case(args.url, case)
        except Exception:  # pylint: disable=broad-exception-caught
            logger.exception('  runtime exception during test case %s (%d)', case_id, case)
            exception_cases.append(case_id)
        logger.setLevel(logging.INFO)
    logger.info('Updating report')
    await update_reports(args.url)
    if exception_cases:
        logger.error('Runtime exception in %d of %d test cases: %s',
                     len(exception_cases), len(test_cases), exception_cases)
        sys.exit(1)


def parse_args() -> argparse.Namespace:
    ''' Parse command line arguments. '''
    parser = argparse.ArgumentParser(description='Autobahn client for'
        ' trio-websocket')
    parser.add_argument('url', help='WebSocket URL for server')
    # TODO: accept case ID's rather than indices
    parser.add_argument('debug_cases', type=int, nargs='*', help='Run'
        ' individual test cases with debug logging (optional)')
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    trio.run(run_tests, args)



import pathlib
import sys

import trustme

def main() -> None:
    here = pathlib.Path(__file__).parent
    ca_path = here / 'fake.ca.pem'
    server_path = here / 'fake.server.pem'
    if ca_path.exists() and server_path.exists():
        print('The CA ceritificate and server certificate already exist.')
        sys.exit(1)
    print('Creating self-signed certificate for localhost/127.0.0.1:')
    ca_cert = trustme.CA()
    ca_cert.cert_pem.write_to_path(ca_path)
    print(f' * CA certificate: {ca_path}')
    server_cert = ca_cert.issue_server_cert('localhost', '127.0.0.1')
    server_cert.private_key_and_cert_chain_pem.write_to_path(server_path)
    print(f' * Server certificate: {server_path}')
    print('Done')


if __name__ == '__main__':
    main()



'''
This simple WebSocket server responds to text messages by reversing each
message string and sending it back.

It also handles ping/pong automatically and will correctly close down a
connection when the client requests it.

To use SSL/TLS: install the `trustme` package from PyPI and run the
`generate-cert.py` script in this directory.
'''
import argparse
import logging

import trio
from trio_websocket import serve_websocket, ConnectionClosed, WebSocketRequest

BIND_IP = '0.0.0.0'
BIND_PORT = 9000
MAX_MESSAGE_SIZE = 16 * 1024 * 1024
logging.basicConfig()
logger = logging.getLogger('client')
logger.setLevel(logging.INFO)
connection_count = 0


async def main() -> None:
    ''' Main entry point. '''
    logger.info('Starting websocket server on ws://%s:%d', BIND_IP, BIND_PORT)
    await serve_websocket(handler, BIND_IP, BIND_PORT, ssl_context=None,
                          max_message_size=MAX_MESSAGE_SIZE)


async def handler(request: WebSocketRequest) -> None:
    ''' Reverse incoming websocket messages and send them back. '''
    global connection_count  # pylint: disable=global-statement
    connection_count += 1
    logger.info('Connection #%d', connection_count)
    ws = await request.accept()
    while True:
        try:
            message = await ws.get_message()
            await ws.send_message(message)
        except ConnectionClosed:
            break
        except Exception:  # pylint: disable=broad-exception-caught
            logger.exception('  runtime exception handling connection #%d', connection_count)


def parse_args() -> argparse.Namespace:
    ''' Parse command line arguments. '''
    parser = argparse.ArgumentParser(description='Autobahn server for'
        ' trio-websocket')
    parser.add_argument('-d', '--debug', action='store_true',
        help='WebSocket URL for server')
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    if args.debug:
        logging.getLogger('trio-websocket').setLevel(logging.DEBUG)
    trio.run(main)
